import pathlib
from typing import Optional

from rbx.box import checkers, limits_info, package, state
from rbx.box.code import CommunicationItem, run_communication, run_item
from rbx.box.environment import EnvironmentSandbox, ExecutionConfig, VerificationLevel
from rbx.box.retries import Retrier, get_retrier_config
from rbx.box.schema import CodeItem, Testcase
from rbx.grading import profiling
from rbx.grading.judge.sandbox import SandboxBase
from rbx.grading.limits import Limits
from rbx.grading.steps import (
    DigestOrDest,
    DigestOrSource,
    Evaluation,
    GradingFileInput,
    GradingFileOutput,
    TestcaseIO,
    TestcaseLog,
)
from rbx.utils import model_to_yaml


def get_limits_for_language(
    lang: Optional[str],
    verification: VerificationLevel,
    timelimit_override: Optional[int],
    use_timelimit: bool = True,
) -> Limits:
    limits = limits_info.get_limits(
        lang,
        profile=limits_info.get_active_profile() or 'local',
        verification=verification,
    )
    if timelimit_override is not None:
        limits.time = timelimit_override
    if limits.time is not None and (not use_timelimit or limits.time <= 0):
        limits.time = None
    return limits


async def run_solution_on_testcase(
    solution: CodeItem,
    compiled_digest: str,
    checker_digest: Optional[str],
    testcase: Testcase,
    output_dir: Optional[pathlib.Path] = None,
    interactor_digest: Optional[str] = None,
    testcase_index: int = 0,
    verification: VerificationLevel = VerificationLevel.NONE,
    timelimit_override: Optional[int] = None,
    use_retries: bool = True,
    use_timelimit: bool = True,
    capture_pipes: Optional[bool] = None,
    nruns: int = 0,
    filestem: Optional[str] = None,
    is_stress: bool = False,
) -> Evaluation:
    if interactor_digest is not None:
        return await _run_communication_solution_on_testcase(
            solution,
            compiled_digest,
            interactor_digest,
            checker_digest,
            testcase,
            output_dir,
            testcase_index=testcase_index,
            verification=verification,
            timelimit_override=timelimit_override,
            use_retries=use_retries,
            use_timelimit=use_timelimit,
            capture_pipes=capture_pipes,
            nruns=nruns,
            filestem=filestem,
            is_stress=is_stress,
        )

    async def run_fn(retry_index: int) -> Evaluation:
        actual_sandbox = package.get_singleton_sandbox()

        limits = get_limits_for_language(
            solution.language,
            verification,
            timelimit_override,
            use_timelimit=use_timelimit,
        )
        extra_config = _get_execution_config(limits, actual_sandbox)

        if output_dir is None:
            assert testcase.outputPath is not None
            output_path = testcase.outputPath
        else:
            stem = filestem or testcase.inputPath.stem
            output_path = output_dir / pathlib.PosixPath(stem).with_suffix('.out')
        error_path = output_path.with_suffix('.err')
        log_path = output_path.with_suffix('.log')
        eval_path = output_path.with_suffix('.eval')
        output_path.parent.mkdir(parents=True, exist_ok=True)

        with profiling.PushContext('tasks.run_solution_on_testcase'):
            run_log = await run_item(
                solution,
                DigestOrSource.create(compiled_digest),
                stdin=DigestOrSource.create(testcase.inputPath),
                stdout=DigestOrDest.create(output_path),
                stderr=DigestOrDest.create(error_path),
                extra_config=extra_config,
                retry_index=retry_index,
            )

        if checker_digest is not None:
            with profiling.PushContext('tasks.run_solution_on_testcase.check'):
                checker_result = await checkers.check(
                    checker_digest,
                    run_log,
                    testcase,
                    program_output=output_path,
                )
        else:
            checker_result = checkers.check_with_no_output(run_log)

        eval = Evaluation(
            result=checker_result,
            testcase=TestcaseIO(
                index=testcase_index,
                input=testcase.inputPath,
                output=testcase.outputPath,
            ),
            log=TestcaseLog(
                **(run_log.model_dump() if run_log is not None else {}),
                stdout_absolute_path=output_path.absolute(),
                stderr_absolute_path=error_path.absolute(),
                log_absolute_path=log_path.absolute(),
                eval_absolute_path=eval_path.absolute(),
            ),
        )

        log_path.write_text(model_to_yaml(eval))
        eval_path.write_text(model_to_yaml(eval))
        return eval

    if not use_retries:
        return await run_fn(0)

    retrier = Retrier(get_retrier_config(nruns), is_stress=is_stress)
    return await retrier.repeat(run_fn)


def _get_execution_config(
    limits: Limits,
    actual_sandbox: SandboxBase,
) -> ExecutionConfig:
    sandbox = EnvironmentSandbox()
    sandbox.timeLimit = limits.time
    if limits.isDoubleTL and sandbox.timeLimit is not None:
        # Double TL.
        sandbox.timeLimit = sandbox.timeLimit * 2
    sandbox.wallTimeLimit = sandbox.timeLimit
    if sandbox.timeLimit is not None and actual_sandbox.use_soft_timeout():
        sandbox.wallTimeLimit = sandbox.timeLimit * 2
    sandbox.memoryLimit = limits.memory
    sandbox.fileSizeLimit = limits.output
    return ExecutionConfig(sandbox=sandbox, problemLimits=limits)


async def _run_communication_solution_on_testcase(
    solution: CodeItem,
    compiled_digest: str,
    interactor_digest: str,
    checker_digest: Optional[str],
    testcase: Testcase,
    output_dir: Optional[pathlib.Path] = None,
    testcase_index: int = 0,
    verification: VerificationLevel = VerificationLevel.NONE,
    timelimit_override: Optional[int] = None,
    use_retries: bool = True,
    use_timelimit: bool = True,
    capture_pipes: Optional[bool] = None,
    nruns: int = 0,
    filestem: Optional[str] = None,
    is_stress: bool = False,
) -> Evaluation:
    if capture_pipes is None:
        capture_pipes = state.STATE.debug_logs

    async def run_fn(retry_index: int) -> Evaluation:
        actual_sandbox = package.get_singleton_sandbox()
        interactor_sandbox = package.get_singleton_interactor_sandbox()

        limits = get_limits_for_language(
            solution.language,
            verification,
            timelimit_override,
            use_timelimit=use_timelimit,
        )

        extra_config = _get_execution_config(limits, actual_sandbox)
        interactor_extra_config = _get_execution_config(limits, interactor_sandbox)
        if (
            interactor_extra_config.sandbox is not None
            and interactor_extra_config.sandbox.wallTimeLimit is not None
            and extra_config.sandbox is not None
            and extra_config.sandbox.wallTimeLimit is not None
        ):
            interactor_extra_config.sandbox.wallTimeLimit += (
                extra_config.sandbox.wallTimeLimit
            )
        # TODO: maybe combine wall time limits?

        if output_dir is None:
            assert testcase.outputPath is not None
            output_path = testcase.outputPath
        else:
            stem = filestem or testcase.inputPath.stem
            output_path = output_dir / pathlib.PosixPath(stem).with_suffix('.out')
        solution_error_path = output_path.with_suffix('.sol.err')
        interactor_error_path = output_path.with_suffix('.int.err')
        log_path = output_path.with_suffix('.log')
        eval_path = output_path.with_suffix('.eval')
        output_path.parent.mkdir(parents=True, exist_ok=True)

        interactor_capture_path = (
            output_path.with_suffix('.pin') if capture_pipes else None
        )
        interactor_item = CommunicationItem(
            code=package.get_interactor(),
            executable=DigestOrSource.create(interactor_digest),
            stderr=DigestOrDest.create(interactor_error_path),
            extra_config=interactor_extra_config,
            extra_args='interactor.in interactor.out',
            inputs=[
                GradingFileInput(
                    src=testcase.inputPath,
                    dest=pathlib.PosixPath('interactor.in'),
                )
            ],
            outputs=[
                GradingFileOutput(
                    src=pathlib.PosixPath('interactor.out'),
                    dest=output_path,
                    touch=True,
                )
            ],
            capture=DigestOrDest.create(interactor_capture_path)
            if interactor_capture_path
            else None,
            file_prefix='interactor',
        )
        solution_capture_path = (
            output_path.with_suffix('.pout') if capture_pipes else None
        )
        solution_item = CommunicationItem(
            code=solution,
            executable=DigestOrSource.create(compiled_digest),
            stderr=DigestOrDest.create(solution_error_path),
            extra_config=extra_config,
            capture=DigestOrDest.create(solution_capture_path)
            if solution_capture_path
            else None,
            file_prefix='solution',
        )

        merged_capture_path = output_path.with_suffix('.pio') if capture_pipes else None
        interactor_run_log, run_log = await run_communication(
            interactor=interactor_item,
            solution=solution_item,
            retry_index=retry_index,
            merged_capture=DigestOrDest.create(merged_capture_path)
            if merged_capture_path
            else None,
        )

        checker_result = await checkers.check_communication(
            checker_digest,
            run_log,
            interactor_run_log,
            interactor_error_path,
            testcase,
            output_path,
        )

        eval = Evaluation(
            result=checker_result,
            testcase=TestcaseIO(
                index=testcase_index,
                input=testcase.inputPath,
                output=testcase.outputPath,
            ),
            log=TestcaseLog(
                **(run_log.model_dump() if run_log is not None else {}),
                stdout_absolute_path=output_path.absolute(),
                stderr_absolute_path=solution_error_path.absolute(),
                log_absolute_path=log_path.absolute(),
                eval_absolute_path=eval_path.absolute(),
            ),
        )

        log_path.write_text(model_to_yaml(eval))
        eval_path.write_text(model_to_yaml(eval))
        interactor_log_path = output_path.with_suffix('.int.log')
        interactor_log_path.unlink(missing_ok=True)
        if interactor_run_log is not None:
            interactor_log_path.write_text(model_to_yaml(interactor_run_log))
        solution_log_path = output_path.with_suffix('.sol.log')
        solution_log_path.unlink(missing_ok=True)
        if run_log is not None:
            solution_log_path.write_text(model_to_yaml(run_log))
        return eval

    if not use_retries:
        return await run_fn(0)

    retrier = Retrier(get_retrier_config(nruns), is_stress=is_stress)
    return await retrier.repeat(run_fn)



# flake8: noqa
from typing import Any, Optional, Union
import typing

import random
import lark

LARK_GRAMMAR = r"""
start: args

// A bunch of args
args: (arg (_WS arg)*)?

// Argument shlex
arg: _block | random_hex

// Blocks
_block: (TEXT | _ticked | _expr)+

// Expression
_expr: var | range | select

// Ticked
_ticked: "`" _expr "`"

// Variables
var: "<" RECNAME ">"

// Select
select: "(" select_value ("|" select_value)* ")"
select_value: _block

// Ranges
range: "[" range_value ".." range_value "]"
range_value: range | var | int | float | char
int: SIGNED_INT
float: SIGNED_FLOAT
char: "'" /./ "'"

// Random hex
random_hex.1: "@"

// Rest, strings, etc
TEXT: (/[^ \t\f\r\n\[\]\(\)\<\>\|\`]/ | ESCAPED_STRING)+

// Whitespace
_WS: WS

RECNAME: /[a-zA-Z0-9_]/+ /(\.[a-zA-Z0-9_])/*

%import common.WS
%import common.CNAME
%import common.SIGNED_INT
%import common.SIGNED_FLOAT
%import common.ESCAPED_STRING
"""

LARK_PARSER = lark.Lark(LARK_GRAMMAR)

Primitive = Union[int, float, str]


class GeneratorParsingError(Exception):
    pass


class RandomInt:
    def __init__(self, min: int, max: int):
        self.min = min
        self.max = max

    def get(self) -> int:
        if self.max < self.min:
            raise GeneratorParsingError(
                f'Found int range with invalid bounds [{self.min}..{self.max}].'
            )
        return random.randint(self.min, self.max)


class RandomChar:
    def __init__(self, min: str, max: str):
        self.min = min
        self.max = max

    def get(self) -> str:
        if len(self.min) != 1 or len(self.max) != 1:
            raise GeneratorParsingError(
                f"Found char range with invalid bounds ['{self.min}'..'{self.max}']"
            )
        mn = ord(self.min)
        mx = ord(self.max)
        if mx < mn:
            raise GeneratorParsingError(
                f"Found char range with invalid bounds ['{self.min}'..'{self.max}']"
            )
        return chr(random.randint(mn, mx))


class RandomHex:
    len: int

    def __init__(self, len: int = 8):
        self.len = len

    def get(self) -> str:
        return ''.join(random.choice('0123456789abcdef') for _ in range(self.len))


def parse(args: str) -> lark.ParseTree:
    tree = LARK_PARSER.parse(args)
    (args_root,) = tree.find_data('args')
    return args_root


def _var_as_str(var: Any) -> str:
    if isinstance(var, float):
        return f'{var:.6f}'
    return str(var)


def _down(node: lark.ParseTree) -> Union[lark.Token, lark.ParseTree]:
    return node.children[0]


def _down_tree(node: lark.ParseTree) -> lark.ParseTree:
    downed = _down(node)
    return typing.cast(lark.ParseTree, downed)


def _down_token(node: lark.ParseTree) -> lark.Token:
    downed = _down(node)
    return typing.cast(lark.Token, downed)


def _is_primitive(x: Primitive) -> bool:
    return isinstance(x, (int, float, str))


def _get_casting_type(a: Primitive, b: Primitive) -> Optional[str]:
    if isinstance(a, int) and isinstance(b, int):
        return 'int'
    if isinstance(a, float) or isinstance(b, float):
        if isinstance(a, (int, float)) or isinstance(b, (int, float)):
            return 'float'
    if isinstance(a, str) and isinstance(b, str):
        return 'char'
    return None


class Generator:
    def __init__(self, vars):
        self.vars = vars

    def handle_var(self, expr: lark.ParseTree) -> Any:
        name = typing.cast(lark.Token, _down(expr))
        if name.value not in self.vars:
            raise GeneratorParsingError(
                f'Error parsing generator expression: variable {name.value} is not defined'
            )
        value = self.vars[name.value]
        if not _is_primitive(value):
            raise GeneratorParsingError(
                f'Variable {name.value} has type {type(value)}, which is not supported by the Generator expression parser.'
            )
        return value

    def handle_range_value(self, range_value: lark.ParseTree) -> Primitive:
        tp = _down_tree(range_value)
        if tp.data == 'int':
            return int(_down_token(tp).value)
        if tp.data == 'float':
            return float(_down_token(tp).value)
        if tp.data == 'char':
            return str(_down_token(tp).value)
        if tp.data == 'var':
            return self.handle_var(tp)
        return self.handle_range(tp)

    def handle_range(self, range: lark.ParseTree) -> Primitive:
        value_a, value_b = range.children[:2]

        item_a = self.handle_range_value(typing.cast(lark.ParseTree, value_a))
        item_b = self.handle_range_value(typing.cast(lark.ParseTree, value_b))

        casting_type = _get_casting_type(item_a, item_b)
        if casting_type is None:
            raise GeneratorParsingError(
                f'Types in range are uncompatible: {type(item_a)} != {type(item_b)}'
            )

        if casting_type == 'int':
            return RandomInt(int(item_a), int(item_b)).get()
        if casting_type == 'float':
            return random.uniform(float(item_a), float(item_b))
        if casting_type == 'char':
            return RandomChar(str(item_a), str(item_b)).get()

        raise GeneratorParsingError(
            f'Types in range are not supported: {type(item_a)}, {type(item_b)}'
        )

    def handle_select_value(self, select_value: lark.ParseTree) -> str:
        items = []
        for item in select_value.children:
            items.append(self.handle_block(item))
        return ''.join(items)

    def handle_select(self, select: lark.ParseTree) -> str:
        options = []
        for select_value in select.children:
            options.append(
                self.handle_select_value(typing.cast(lark.ParseTree, select_value))
            )
        return options[RandomInt(0, len(options) - 1).get()]

    def handle_expr(self, expr: lark.ParseTree) -> str:
        if expr.data == 'var':
            return _var_as_str(self.handle_var(expr))
        if expr.data == 'range':
            return _var_as_str(self.handle_range(expr))
        if expr.data == 'select':
            return self.handle_select(expr)
        raise GeneratorParsingError(
            f'Internal error: found invalid AST node in Generator parsing, {expr.data}'
        )

    def handle_block(self, block: Union[lark.ParseTree, lark.Token]) -> str:
        if isinstance(block, lark.Token):
            return block.value

        return self.handle_expr(block)

    def generate_arg(self, arg: lark.ParseTree) -> str:
        items = []
        for arg_item in arg.children:
            if hasattr(arg_item, 'data'):
                data_item = typing.cast(lark.ParseTree, arg_item)
                if data_item.data == 'random_hex':
                    items.append(RandomHex().get())
                    continue
            items.append(self.handle_block(arg_item))
        return ''.join(items)

    def generate(self, args: lark.ParseTree) -> str:
        args_list = []

        for arg in args.children:
            args_list.append(self.generate_arg(typing.cast(lark.ParseTree, arg)))

        return ' '.join(args_list)


if __name__ == '__main__':
    tree = parse(
        """--MAX_N="allow me" --int=[1..<MAX_N>] --float=[1.0..<MAX_N>] --char=['a'..'a'] --selection="(a|b|(c|<MAX_N>))" --select=("a"|"b"|"c") @ --r2=[1..[8..15]]"""
    )
    print(tree)

    generator = Generator({'MAX_N': 10})
    print(generator.generate(tree))



import functools
import pathlib
import shutil

from rbx import console, utils
from rbx.box.schema import CodeItem
from rbx.grading.judge.cacher import FileCacher
from rbx.grading.steps import GradingFileOutput


class WarningStack:
    def __init__(self, root: pathlib.Path):
        self.root = root
        self.warnings = set()
        self.sanitizer_warnings = {}

    def add_warning(self, code: CodeItem):
        self.warnings.add(code.path)

    def add_sanitizer_warning(
        self, cacher: FileCacher, code: CodeItem, reference: GradingFileOutput
    ):
        if code.path in self.sanitizer_warnings:
            return
        dest_path = _get_warning_runs_dir(self.root).joinpath(
            code.path.with_suffix(code.path.suffix + '.log')
        )
        dest_path.parent.mkdir(parents=True, exist_ok=True)
        f = reference.get_file(cacher)
        if f is None:
            return
        with dest_path.open('wb') as fout:
            shutil.copyfileobj(f, fout)
        f.close()
        self.sanitizer_warnings[code.path] = dest_path

    def clear(self):
        self.warnings.clear()
        self.sanitizer_warnings.clear()


@functools.cache
def _get_warning_stack(root: pathlib.Path) -> WarningStack:
    return WarningStack(root)


@functools.cache
def _get_cache_dir(root: pathlib.Path) -> pathlib.Path:
    dir = root / '.box'
    dir.mkdir(parents=True, exist_ok=True)
    return dir


@functools.cache
def _get_warning_runs_dir(root: pathlib.Path) -> pathlib.Path:
    dir = _get_cache_dir(root) / 'warnings'
    shutil.rmtree(dir, ignore_errors=True)
    dir.mkdir(parents=True, exist_ok=True)
    return dir


def get_warning_stack() -> WarningStack:
    current_root = utils.abspath(pathlib.Path.cwd())
    return _get_warning_stack(current_root)


def print_warning_stack_report():
    stack = get_warning_stack()
    if not stack.warnings and not stack.sanitizer_warnings:
        return
    console.console.rule('[status]Warning stack[/status]')
    console.console.print(
        f'[warning]There were some warnings within the code that run at [item]{stack.root.absolute()}[/item][/warning]'
    )
    if stack.warnings:
        console.console.print(f'{len(stack.warnings)} compilation warnings')
        console.console.print(
            'You can use [item]rbx compile[/item] to reproduce the issues with the files below.'
        )
        for path in sorted(stack.warnings):
            console.console.print(f'- [item]{path}[/item]')
        console.console.print()

    if stack.sanitizer_warnings:
        console.console.print(f'{len(stack.sanitizer_warnings)} sanitizer warnings')
        for path in sorted(stack.sanitizer_warnings):
            console.console.print(
                f'- [item]{path}[/item], example log at [item]{stack.sanitizer_warnings[path]}[/item]'
            )
        console.console.print()
